import ast
from io import StringIO
from tokenize import generate_tokens
from typing import Optional

import pytest

from flake8_hangover import Plugin

CLASSES_REGISTRY = {}


@pytest.fixture
def run_plugin():
    """Fixture to parse ast from string and run plugin on it."""
    def wrapper(code_str, strip_tabs: Optional[int] = None):
        if strip_tabs:
            tabs = strip_tabs * 4
            code_str = '\n'.join([line[tabs:] for line in code_str.strip('\n').split('\n')])
        tree = ast.parse(code_str)
        tokens = list(generate_tokens(StringIO(code_str).readline))
        return {'{}:{}: {}'.format(*r) for r in Plugin(tree=tree, file_tokens=tokens).run()}
    return wrapper


def register_case(_class):
    """Decorator to register case class."""
    name = _class.__name__
    module = _class.__module__
    if module not in CLASSES_REGISTRY:
        CLASSES_REGISTRY[module] = {}
    if name in CLASSES_REGISTRY[module]:
        raise ValueError(f'Case "{name}" already exists')
    CLASSES_REGISTRY[module][name] = _class
    return _class


def check_registered_case(
    run_plugin,
    case,
    case_name: str = None,
    case_code: str = None,
):
    """Test plugin on function calls."""
    name = case_name or case.__name__
    code = case_code or case.code
    expected_errors = sorted(case.errors or [])
    found_errors = sorted(list(run_plugin(code, strip_tabs=1)), key=lambda s: s.split(': ', 1)[1])

    def show_error(found, expected):
        if isinstance(found, list):
            found = '\n'.join(found)
        if isinstance(expected, list):
            expected = '\n'.join(expected)
        return (
            f'Case "{name}" failed.\n'
            f'Code:\n{code}\n'
            f'Errors:\n{found}\n\n'
            f'Expected:\n{expected}\n'
        )

    if expected_errors:
        assert len(found_errors) == len(expected_errors), show_error(found_errors, expected_errors)
        for i, msg in enumerate(expected_errors):
            assert found_errors[i].endswith(msg), show_error(found_errors, expected_errors)
    else:
        assert not found_errors, show_error(found_errors, 'No errors')



import ast
import tokenize
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
)

from .__version__ import __version__
from .messages import Messages
from .validator import IndentValidator

TAB_SIZE = 4


class Visitor(ast.NodeVisitor):
    """Class for visiting ast nodes."""

    def __init__(self, tokens: List[tokenize.TokenInfo]) -> None:
        """Initialize class instance."""
        self.errors: Dict[Tuple[int, int], str] = {}
        self._tokens = tokens

    def add_error(self, lineno: int, offset: int, error: str) -> None:
        """Add error (unique only) to errors list."""
        key = (lineno, offset)
        if key not in self.errors:
            self.errors[key] = error

    def visit_Call(self, node: ast.Call) -> None:
        """Visit ``Call`` node."""
        cur_lineno = node.lineno
        func_name_offset = None
        last_inner_lineno = cur_lineno  # not include args which started at the same line as node

        # Iterate over positional arguments
        for arg in node.args:
            arg_col_offset = self._get_arg_col_offset(arg)
            arg_lineno = self._get_arg_lineno(arg)

            if arg_lineno - cur_lineno == 1:
                if func_name_offset is None:
                    func_name_offset = self._get_func_name_offset(node)

                if arg_col_offset > func_name_offset or arg_col_offset % TAB_SIZE != 0:
                    self.add_error(arg_lineno, arg_col_offset, Messages.FHG002)

            cur_lineno = self._get_arg_end_lineno(arg, default=arg_lineno)
            if arg_lineno != node.lineno:
                last_inner_lineno = max(last_inner_lineno, cur_lineno)

        # Iterate over keyword arguments
        for kwarg in node.keywords:
            kwarg_col_offset = self._get_arg_col_offset(kwarg)
            kwarg_lineno = self._get_arg_lineno(kwarg)

            if kwarg_lineno - cur_lineno == 1:
                if func_name_offset is None:
                    func_name_offset = self._get_func_name_offset(node)

                if (
                    kwarg_col_offset > func_name_offset
                    or (kwarg.arg and kwarg_col_offset % TAB_SIZE != 0)
                ):
                    self.add_error(kwarg_lineno, kwarg_col_offset, Messages.FHG003)

            cur_lineno = self._get_arg_end_lineno(kwarg, default=kwarg_lineno)
            if getattr(kwarg, 'lineno', kwarg_lineno) != node.lineno:
                last_inner_lineno = max(last_inner_lineno, cur_lineno)

        self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        """Visit ``FunctionDef`` node."""
        self._check_func_args_indentations(node)
        self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        """Visit ``AsyncFunctionDef`` node."""
        self._check_func_args_indentations(node)
        self.generic_visit(node)

    def _check_func_args_indentations(self, node: Any) -> None:
        """Check indentations in function args/kwargs."""
        cur_lineno = node.lineno
        first_argument = None
        multiline_arguments = False

        for i, arg in enumerate(node.args.args):
            if i == 0:
                first_argument = (arg.lineno, arg.col_offset)

            if arg.lineno != cur_lineno:
                if arg.col_offset != node.col_offset + 4:
                    self.add_error(arg.lineno, arg.col_offset, Messages.FHG001)
                cur_lineno = arg.lineno
                multiline_arguments = True

        if (
            multiline_arguments
            and first_argument
            and first_argument[0] == node.lineno
        ):
            self.add_error(*first_argument, Messages.FHG004)

    def _get_arg_col_offset(self, obj: Any) -> int:
        """Get `col_offset` for object."""
        if isinstance(obj, ast.keyword):
            return obj.value.col_offset - len(str(obj.arg or '')) - 1  # 1 is for "="
        if isinstance(obj, ast.GeneratorExp):
            return self._get_arg_col_offset(obj.elt)
        return int(obj.col_offset)

    def _get_arg_lineno(self, obj: Any) -> int:
        """Get `lineno` for object."""
        if isinstance(obj, ast.keyword):
            return self._get_arg_lineno(obj.value)
        if isinstance(obj, ast.GeneratorExp):
            return self._get_arg_lineno(obj.elt)

        return int(obj.lineno)

    def _get_arg_end_lineno(self, obj: Any, default: Optional[int] = None) -> int:
        """Get `end_lineno` for object."""
        if isinstance(obj, ast.GeneratorExp):
            last_gen = obj.generators[-1]
            return max(
                self._get_arg_end_lineno(last_gen.target),
                self._get_arg_end_lineno(last_gen.iter),
            )

        return getattr(obj, 'end_lineno', int(obj.lineno if default is None else default))

    def _get_func_name_offset(self, node: Any) -> int:
        """Get function name offset."""
        func_name = self._get_func_name(node.func)
        return int(node.col_offset + max(len(func_name), TAB_SIZE))

    def _get_func_name(self, obj: Any) -> str:
        """Extract function full name from node.

        May not fully correctly work. For this cases function returns empty string.
        """
        try:
            if isinstance(obj, ast.Attribute):
                return f'{self._get_func_name(obj.value)}.{obj.attr}'
            if isinstance(obj, ast.Call):
                return self._get_func_name(obj.func)
            if isinstance(obj, ast.Subscript):
                return f'{self._get_func_name(obj.value)}[{self._get_func_name(obj.slice)}]'
            if isinstance(obj, ast.Index):
                return self._get_func_name(obj.value)
            if isinstance(obj, ast.Constant):
                return str(obj.value)

            return str(obj.id)
        except Exception:
            return ''


class Plugin:
    """Class to run flake8 plugin."""

    name = 'flake8-hangover'
    version = __version__

    def __init__(self, tree: ast.AST, file_tokens: List[tokenize.TokenInfo]):
        """Initialize class instance."""
        self._tree = tree
        self._tokens = file_tokens

    def run(self) -> Generator[Tuple[int, int, str, Type[Any]], None, None]:
        """Run plugin."""
        visitor = Visitor(tokens=self._tokens)
        visitor.visit(self._tree)

        for error_key, error_msg in visitor.errors.items():
            lineno, col_offset = error_key
            yield lineno, col_offset, error_msg, type(self)

        indent_validator = IndentValidator(tokens=self._tokens)
        indent_validator.validate()
        for error_key, error_msg in indent_validator.errors.items():
            lineno, col_offset = error_key
            yield lineno, col_offset, error_msg, type(self)



from queue import LifoQueue
from token import (
    DEDENT,
    INDENT,
    OP,
)
from tokenize import TokenInfo
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

from .messages import Messages


class Parenthese:
    """Store full info for single parentheses."""

    def __init__(
        self,
        line: int,
        column: int,
        type: int,
        open: bool,
        indent: int = 0,
        pair: Optional['Parenthese'] = None,
    ) -> None:
        self.line = line
        self.column = column
        self.type = type
        self.open = open
        self.indent = indent
        self.pair = pair

    @classmethod
    def parse_token(
        cls, token: TokenInfo, indents: Dict[int, int],
    ) -> Optional['Parenthese']:
        """Create class instance from token."""
        if token.type != OP or token.string not in '()[]{}':
            return None
        return cls(
            line=token.start[0],
            column=token.start[1],
            type=cls.get_type(token.string),
            open=token.string in '([{',
            indent=indents[token.start[0]],
        )

    @classmethod
    def get_type(cls, s: str) -> int:
        """Split tokens by type."""
        types = {'(': 1, ')': 1, '[': 2, ']': 2, '{': 3, '}': 3}
        return types.get(s, 0)

    def __repr__(self) -> str:
        """Object representation."""
        s = f"{self.line}:{self.column} ({self.indent}) type:{self.type} open:{self.open}"
        if self.pair:
            s += f" | pair: <{self.pair.line}:{self.pair.column} ({self.pair.indent})>"
        return s


class IndentValidator:
    """Validate close parentheses have the same line indent as open ones."""

    def __init__(self, tokens: List[TokenInfo]) -> None:
        self.tokens = tokens
        self.errors: Dict[Tuple[int, int], str] = {}

    def validate(self) -> None:
        """Check all parentheses"""
        indents = self.calculate_indents(self.tokens)
        parentheses = self.parse_parentheses(self.tokens, indents)
        for p in parentheses:
            if not p.open:
                continue
            if not p.pair or p.line == p.pair.line:
                continue
            if p.indent != p.pair.indent:
                self.errors[(p.pair.line, p.pair.column)] = Messages.FHG005

    def calculate_indents(self, tokens: List[TokenInfo]) -> Dict[int, int]:
        """Calculate the indent for each line in the code."""
        indents = {}
        for t in tokens:
            if t.type in (INDENT, DEDENT):
                continue
            line = t.start[0]
            if line not in indents:
                indents[line] = t.start[1]
        return indents

    def parse_parentheses(
        self,
        tokens: List[TokenInfo],
        indents: Dict[int, int],
    ) -> List[Parenthese]:
        """Parse all parentheses with their positions."""
        parentheses: List[Parenthese] = []
        # find all parentheses
        for t in tokens:
            p = Parenthese.parse_token(t, indents)
            if not p:
                continue
            parentheses.append(p)
        # find pairs for each parentheses
        queues: Dict[int, LifoQueue] = {1: LifoQueue(), 2: LifoQueue(), 3: LifoQueue()}
        for p in parentheses:
            if p.open:
                queues[p.type].put(p)
            else:
                if queues[p.type].empty():
                    raise ValueError(f"Unexpected close parentheses {p}")
                p_open = queues[p.type].get()
                p.pair = p_open
                p_open.pair = p
                assert p.open != p_open.open
                assert p.type == p_open.type
        return parentheses



import pytest

from flake8_hangover.plugin import Messages

from .conftest import (
    CLASSES_REGISTRY,
    check_registered_case,
    register_case,
)


@register_case
class Case1:
    errors = [Messages.FHG005]
    code = """
    predicted: list = ([predicted_columns] if isinstance(predicted_columns, str)
                    else predicted_columns)
    """


@register_case
class Case2:
    errors = [Messages.FHG005]
    code = """
    predicted = ([predicted_columns] if isinstance(predicted_columns, str)
                    else predicted_columns
        )
    """


@register_case
class Case3:
    errors = None
    code = """
    predicted: list = ([predicted_columns] if isinstance(predicted_columns, str)
                    else predicted_columns
    )
    """


@register_case
class Case4:
    errors = None
    code = """
    predicted = (
        [predicted_columns] if isinstance(predicted_columns, str)
        else predicted_columns
    )
    """


@register_case
class Case5:
    errors = None
    code = """
    value = func(
        param=1,
    ).arg()
    """


@register_case
class Case6:
    errors = None
    code = """
    value = model.score(
        param=1,
    ) * 2 - 1
    """


@register_case
class Case7:
    errors = None
    code = """
    final_score['index'] = [
        i for i in range(0, len(sorted_columns), self.step)
    ][:final_score.shape[0]]
    """


@register_case
class Case8:
    errors = None
    code = """
    data_to_plot = create_distplot(
        bin_size=bin_size,
    ).data
    """


@register_case
class Case9:
    errors = None
    code = """
    class MyClass:
        some_results: Union[
            int,
            HyperparametersOptimizationResult,
            ModelInterpretationResult,
        ] = 100
    """


@register_case
class Case10:
    errors = None
    code = """
    n_best_scores = sorted(
        deserialized_scores,
    )[-max_scores_on_plot:]
    """


@pytest.mark.parametrize('case', CLASSES_REGISTRY[__name__].values())
def test_plugin_simple_indent(run_plugin, case):
    """Test plugin on simple indentation."""
    check_registered_case(run_plugin, case)



"""
Tests for following rules:
    - FHG002 Function call positional argument has hanging indentation
    - FHG003 Function call keyword argument has hanging indentation
"""
import pytest

from flake8_hangover.plugin import Messages

from .conftest import (
    CLASSES_REGISTRY,
    check_registered_case,
    register_case,
)


@register_case
class Case1:
    errors = None
    code = """
    def foo():
        if use_shap:
            shap_values, shap_interaction_values = _calc_shap(
                df, estimator, feature_columns, shap_interactions,
            )
    """


@register_case
class Case2:
    errors = None
    code = """
    def foo():
        if use_shap:
            shap_values, shap_interaction_values = _calc_shap(
                df,
                estimator,
                feature_columns,
                shap_interactions,
            )
    """


@register_case
class Case3:
    errors = [Messages.FHG005]
    code = """
    def foo():
        my_func(value='name',
            other_value='hello')
    """


@register_case
class Case4:
    errors = None
    code = """
    print(  # noqa
        'one line'
        f'{report_dir_path / report_name}', end=' ',
    )
    """


@register_case
class Case5:
    errors = None
    code = """
    try:
        current_controller = CacheController(
            check_version_level=check_version_level,
            check_func_code_hash=check_func_code_hash,
            **controller_kwargs,
        )
    except ValidationError as e:
        raise ValueError()
    """


@register_case
class Case6:
    errors = None
    code = """
    current_controller = CacheController(
        *controller_kwargs,
    )
    """


@register_case
class Case7:
    errors = None
    code = """
    t_group, bins_data = get_groups(
        df=df[feature_columns + [target_column]],
        continuous=continuous,
        categorical=categorical,
        target_column=target_column,
        **(grouping_params or {}),
    )
    """


@register_case
class Case8:
    errors = None
    code = """
    return cls(
        func_code_hash=fixed_hash or joblib.hash(
            func.__code__.co_consts
            + (func.__code__.co_code,),
        ),
        func_args_kwargs={
            'args': yaml_repr(func_args or []),
            'kwargs': yaml_repr(func_kwargs or {}),
        },
        **base_meta.dict(),
    )
    """


@register_case
class Case9:
    errors = None
    code = """
    subplot = Plot(
        linewidth=(
            wide_line
            if best_score and score.params == best_score.params
            else default_line
        ),
    )
    """


@register_case
class Case10:
    errors = None
    code = """
    subprocess.run(
        f'python{python_version} -m venv venv',
        shell=True, check=True, cwd=directory,
    )
    """


@register_case
class Case11:
    errors = None
    code = """
    x(
        shell=True,
        check=True,
        cwd=directory,
    )
    """


@register_case
class Case12:
    errors = None
    code = """
    x(
        1,
        2,
        y=10,
    )
    """


@register_case
class Case13:
    errors = None
    code = """
    for key in predictor_instance.dict().keys():
        value = getattr(predictor_instance, key)
    """


@register_case
class Case14:
    errors = None
    code = """
    kafka_config = {'bootstrap.servers': settings["kafka"].get("url")}
    """


@register_case
class Case15:
    errors = None
    code = """
    def _format_error(self, value, message) -> str:
        return (self.error or message).format(
            input=value,
            min=self.format_min(),
            max=self.format_max(),
        )
    """


@register_case
class Case16:
    errors = None
    code = """
    if a != b:
        error_message = get_error_message(param,
            other_param,
        )
    """


@register_case
class Case17:
    errors = None
    code = """
    if a != b:
        error_message = get_error_message(
            param,
            other_param,
        )
    """


@register_case
class Case18:
    errors = [Messages.FHG002, Messages.FHG005]
    code = """
    def foo():
        if use_shap:
            shap_values, shap_interaction_values = _calc_shap(df, estimator, feature_columns,
                                                                shap_interactions)
    """


@register_case
class Case19:
    errors = [Messages.FHG002, Messages.FHG005]
    code = """
    if a != b:
        error_message = get_error_message(param,
                                          other_param)
    """


@register_case
class Case20:
    errors = [Messages.FHG003, Messages.FHG005]
    code = """
    def foo():
        my_func(value='name',
                other_value='hello')
    """


@register_case
class Case21:
    errors = [Messages.FHG005]
    code = """
    def foo():
        my_func(
            value='name',
            other_value='hello')
    """


@register_case
class Case22:
    errors = [Messages.FHG005]
    code = """
    def foo():
        my_func(
            'name',
            'hello')
    """


@register_case
class Case23:
    errors = [Messages.FHG005]
    code = """
    def foo():
        my_func(
            value='name',
            other_value='hello'
            )
    """


@register_case
class Case24:
    errors = [Messages.FHG005]
    code = """
    def foo():
        my_func(
            'name',
            'hello'
            )
    """


@register_case
class Case25:
    errors = [Messages.FHG005]
    code = """
    def foo():
        result = my_func(
            value='name',
            other_value='hello'
            )
    """


@register_case
class Case26:
    errors = None
    code = """
    def foo():
        my_func({(
            'name',
            'hello'
        )})
    """


@register_case
class Case27:
    errors = None
    code = """
    def foo():
        my_func({(
            'name'
        )})
    """


@register_case
class Case28:
    errors = None
    code = """
    def foo():
        my_func({(  # comment with brackets ((
            123
        )})
    """


@register_case
class Case29:
    errors = None
    code = '''
    my_str = """
    hey
    """
    '''


@register_case
class Case30:
    errors = None
    code = """
    hello = '''
    world
    '''
    """


@register_case
class Case31:
    errors = None
    code = """
    @parametrize('value, error_messages', (
        (incorrect_1, None),
    ))
    def my_func():
        pass
    """


@register_case
class Case32:
    errors = None
    code = """
    def shorten_key(value: Any) -> str:
        return divider.join(
            part[:max_part_length]
            for part in value
        )
    """


@register_case
class Case33:
    errors = [Messages.FHG005]
    code = """
    def shorten_key(value: Any) -> str:
        return divider.join(
            part[:max_part_length]
            for part in value)
    """


@register_case
class Case34:
    errors = None
    code = """
    subplot = Plot(
        x=list(range(1, len(score.cv_scores) + 1)),  # number of CV split on X axis
        y=score.cv_scores,  # CV score on Y axis
        label=score.tag or f'#{i}' if enable_labels else None,
        linewidth=(
            wide_line
            if best_score and score.params == best_score.params
            else default_line
        ),
        set_xticks=True,
        plot_type=plot_type,
    )
    """


@register_case
class Case35:
    errors = None
    code = """
    return str(plot_subplots(
        subplots=[plot],
    ))
    """


@register_case
class Case36:
    errors = None
    code = """
    shap_list = np.abs(
        param=10,
    ).mean(0)
    """


@register_case
class Case37:
    errors = None
    code = """
    self.X_val[sorted_columns].columns[
        sorted_importances_idx[::-1]
    ].to_list()
    """


@register_case
class Case38:
    errors = None
    code = """
    result = [Class(
        kw='',
    )]
    """


@pytest.mark.parametrize('case', CLASSES_REGISTRY[__name__].values())
def test_plugin_on_func_call(run_plugin, case):
    """Test plugin on function calls."""
    check_registered_case(run_plugin, case)



class Messages:
    """Linter messages."""

    FHG001 = 'FHG001 Function argument has hanging indentation'
    FHG002 = 'FHG002 Function call positional argument has hanging indentation'
    FHG003 = 'FHG003 Function call keyword argument has hanging indentation'
    FHG004 = 'FHG004 First function argument must be on new line'
    FHG005 = 'FHG005 Close bracket have different indentation with open bracket'



"""
Tests for following rules:
    - FHG001 Function argument has hanging indentation
"""
import pytest

from flake8_hangover.plugin import Messages

from .conftest import (
    CLASSES_REGISTRY,
    check_registered_case,
    register_case,
)


@register_case
class Case1:
    errors = None
    code = """
    def _calc_pdp(
        df: pd.DataFrame, estimator: sklearn.base.BaseEstimator, feature_columns: List[str],
        pdp_kwarg: Optional[Dict] = None,
    ) -> List[pdp.PDPIsolate]:
        pdp_isolates = []
    """


@register_case
class Case2:
    errors = None
    code = """
    def _calc_pdp(
        df: pd.DataFrame,
        estimator: sklearn.base.BaseEstimator,
        feature_columns: List[str],
        pdp_kwarg: Optional[Dict] = None,
    ) -> List[pdp.PDPIsolate]:
        pdp_isolates = []
    """


@register_case
class Case3:
    errors = None
    code = """
    def _calc_pdp(df: pd.DataFrame, estimator: sklearn.base.BaseEstimator) -> List[pdp.PDPIsolate]:
        pdp_isolates = []
    """


@register_case
class Case4:
    errors = None
    code = """
    def _calc_pdp(
        df: pd.DataFrame,
        estimator: sklearn.base.BaseEstimator,
        feature_columns: List[str], pdp_kwarg: Optional[Dict] = None,
    ) -> List[pdp.PDPIsolate]:
        pdp_isolates = []
    """


@register_case
class Case5:
    errors = None
    code = """
    def _hello_world(
        param: pd.DataFrame,
        other_param: sklearn.base.BaseEstimator,
        extra_param: Optional[Dict] = None,
    ) -> str:
        ...
    """


@register_case
class Case6:
    errors = [Messages.FHG004, Messages.FHG005]
    code = """
    def _hello_world(param: pd.DataFrame, other_param: sklearn.base.BaseEstimator,
        extra_param: Optional[Dict] = None) -> str:
        ...
    """


@register_case
class Case7:
    errors = [Messages.FHG004, Messages.FHG001, Messages.FHG005]
    code = """
    def _calc_pdp(df: pd.DataFrame, estimator: sklearn.base.BaseEstimator,
                pdp_kwarg: Optional[Dict]) -> List[pdp.PDPIsolate]:
        pdp_isolates = []
    """


@register_case
class Case8:
    errors = [Messages.FHG004, Messages.FHG001, Messages.FHG005]
    code = """
    def _hello_world(param: pd.DataFrame, other_param: sklearn.base.BaseEstimator,
                    extra_param: Optional[Dict] = None) -> str:
        ...
    """


@register_case
class Case9:
    errors = [Messages.FHG001] * 3
    code = """
    def _calc_pdp(
            df: pd.DataFrame,
            estimator: sklearn.base.BaseEstimator,
            feature_columns: List[str], pdp_kwarg: Optional[Dict] = None,
    ) -> List[pdp.PDPIsolate]:
        pdp_isolates = []
    """


@register_case
class Case10:
    errors = [Messages.FHG004, Messages.FHG005]
    code = """
    def test_something(foo, bar,
        buzz):
        return None
    """


@register_case
class Case11:
    errors = [Messages.FHG004]
    code = """
    def test_something(key=foo,
        value=buzz,
    ):
        return key + value
    """


@register_case
class Case12:
    errors = None
    code = """
    def test_something(
        foo, bar, buzz,
    ):
        return None
    """


@pytest.mark.parametrize('case', CLASSES_REGISTRY[__name__].values())
def test_plugin_on_func_definition(run_plugin, case):
    """Test plugin on function definitions."""
    check_registered_case(run_plugin, case)


@pytest.mark.parametrize('case', CLASSES_REGISTRY[__name__].values())
def test_plugin_on_async_func_definition(run_plugin, case):
    """Test plugin on async function definitions."""
    check_registered_case(
        run_plugin,
        case,
        case_name=f'ASYNC {case.__name__}',
        case_code=case.code.replace('def ', 'async def '),
    )
